/* Zadatak 3A:

Podatke o restoranima u jednoj ulici spremamo u vezanu listu. Za svaki restoran
pamtimo naziv (string od najvie 50 znakova), ku\'{c}ni broj (prirodni broj) i
zaradu restorana (realni broj).
Vezana lista restorana je uvijek \textit{uzlazno\/} sortirana po ku\'{c}nom broju.
Moete pretpostaviti da ne postoje dva restorana s istim ku\'{c}nim brojem.
\begin{itemize}
  \item[(a)]
	Napiite definiciju odgovarajuceg tipa podatka za restoran, na nacin
	da bude moguce definirati varijable naredbom \texttt{rest a;} i sl.
	Definirajte samo podatke koji su nuni za cuvanje takve liste u memoriji.
  \item[(b)]
	Napiite funkciju \texttt{rest* unesi(rest* first, rest* new)} koja
	prima vezanu listu restorana (\texttt{first} je pokazivac na pocetak
	ulazne vezane liste) te restoran \texttt{new}. Funkcija ubacuje element
	\texttt{new} u vezanu listu tako da vezana lista i dalje bude
	\textit{uzlazno\/} sortirana po ku\'{c}nom broju. Ulazna vezana lista moe
	biti prazna. Funkcija vraca pokazivac na prvi element dobivene vezane
	liste, nakon ubacivanja novog elementa.
  \item[(c)]
	Napiite funkciju \texttt{rest* prvih\_k(rest** first, int k)} koja
	trai prvih \texttt{k} susjednih restorana s ukupnom najvecom zaradom
	i izbacuje ih iz vezane liste. Dva restorana su susjedna ako u vezanoj
	listi ne postoji restoran koji se nalazi izmedu njih.
	Funkcija vraca pokazivac na pocetak vezane liste izbacenih restorana
	(koja, takoder, mora biti sortirana \textit{uzlazno\/} po ku\'{c}nim
	brojevima). Ukoliko u ulaznoj vezanoj listi ima strogo manje od
	\texttt{k} elemenata, funkcija ne mijenja ulaznu vezanu listu i vraca
	\texttt{NULL}\@.
  \item[(d)]
	U glavnom programu formirajte vezanu listu od $5$ restorana (elemenata),
	koristeci funkciju \texttt{unesi}. Podatke za te restorane mo\v{z}ete
	u\v{c}itati, ili zadati kao neke konstantne vrijednosti.
	Zatim, pozovite funkciju \texttt{prvih\_k} za $k = 3$. Na kraju,
	dealocirajte memoriju za sve elemente.
	Za rjeavanje ovog podzadatka nije nuno rijeiti podzadatke (b) i (c).
\end{itemize}
U (b) i (c) nije dozvoljeno alocirati memoriju za nove elemente, vec sve treba
raditi s postojecim elementima.

*/

#include <stdio.h>
#include <stdlib.h>

    /* Deklaracija tipa iz (a). */

typedef struct _restoran {
    char naziv[51];
    int kucni_broj;
    double zarada;
    struct _restoran *next;
} restoran;

    /* Funkcija unesi iz (b). */

restoran* unesi(restoran* first, restoran* new)
{
    restoran *tmp;

	/* Slucaj kad je ulazna lista prazna. */
    if (first == NULL) {
        new->next = NULL;
        return new;
    }

        /* Slucaj kad novi element dodajemo na pocetak
           pa se mijenja prvi element liste. */
    if (new->kucni_broj < first->kucni_broj) {
        new->next = first;
        return new;
    }

        /* Ostali slucajevi = novi element dolazi iza
           nekog elementa u listi (first se NE mijenja). */
        /* Prazna for petlja koja staje u trenutku kad
           new treba doci izmedju tmp i tmp->next. */
    for (tmp = first; tmp->next != NULL && new->kucni_broj > tmp->next->kucni_broj; tmp = tmp->next);
    new->next = tmp->next;
    tmp->next = new;

    return first;
}

    /* Funkcija prvih_k iz (c). */
    /* Napomena: Pretpostavljamo da je k >= 1 (duljina trazenog bloka)! */

restoran* prvih_k(restoran** first, int k)
{
        /* Pokazivac na element ispred maksimalnih k, osim ako je pocetak
           tog bloka bas na pocetku liste, tj. nema onog ispred bloka. */
    restoran *maxfirst;
        /* Pokazivac na zadnji element u bloku od maksimalnih k. */
    restoran *maxlast;
        /* Analogni pokazivaci na blok od k elemenata kojeg provjeravamo,
           ako u listi ima vise takvih blokova (lista ima > k elemenata). */
    restoran *tmpfirst, *tmplast;

    int n = 0;
    int firstmax;  // Logicka varijabla = je li max. blok na pocetku liste.
    double maxzarada = 0.0, tmpzarada;

	/* Prvih k elemenata u listi uzimamo kao pocetni maksimum i
           pazimo da zaista imamo barem k elemenata u ulaznoj listi. */
    maxfirst = *first;
    for (maxlast = maxfirst; maxlast != NULL; maxlast = maxlast->next) {
        maxzarada += maxlast->zarada;
        ++n;
        if (n == k)
            break;    // Lista ima (barem) k elemenata.
    }

	/* Ako ulazna lista ima strogo manje od k elemenata,
           onda nista ne mijenjamo i vratimo NULL. */
    if (n < k)
        return NULL;

        /* Ulazna lista ima barem k elemenata, a maxlast pokazuje bas na
           k-ti element u listi (zbog break, maxlast se nije pomaknuo).
           Trenutni max. blok od k elemenata je na pocetku liste. */
    firstmax = 1;

        /* Provjeri ostale blokove od susjednih k elemenata (ako ih ima). */
        /* Pokazivac tmpfirst pokazuje na element ispred bloka od k susjednih
           elemenata, a tmplast pokazuje na zadnji element u tom bloku. */

    tmpfirst = maxfirst;        // Ispred sljedeceg bloka, ako ga ima.
    tmplast = maxlast->next;    // Zadnji u sljedecem bloku, ako ga ima.
    tmpzarada = maxzarada;

    while (tmplast != NULL) {
            /* Popravak zarade za sljedeci blok od k elemenata. */
        tmpzarada += tmplast->zarada - tmpfirst->zarada;
            /* Ako je to nova maksimalna, popravi sve podatke. */
        if (tmpzarada > maxzarada) {
            maxzarada = tmpzarada;
            maxfirst = tmpfirst;
            maxlast = tmplast;
            firstmax = 0;    // Max. blok vise nije na pocetku liste.
        }
        tmpfirst = tmpfirst->next;
        tmplast = tmplast->next;
    }

    if (firstmax) {
            /* Izbaci blok s pocetka liste. */
        *first = maxlast->next;
        maxlast->next = NULL;
        return maxfirst;
    }
    else {
            /* Izbaci blok iza maxfirst, do maxlast (ukljucivo). */
        tmpfirst = maxfirst->next;    // Zapamti pocetak izbacenog dijela.
        maxfirst->next = maxlast->next;
        maxlast->next = NULL;         // Kraj izbacenog dijela.
        return tmpfirst;
    }
}

    /* Pomocna funkcija za ispis elemenata liste. */

void ispis(restoran *first)
{
    restoran *tmp;

    for (tmp = first; tmp != NULL; tmp = tmp->next) {
        printf("%s\n%d, %f\n\n", tmp->naziv, tmp->kucni_broj, tmp->zarada);
    }
    printf("\n");

    return;
}


    /* Glavni program iz (d). */

int main(void)
{
    int n = 5, k = 3, i;
    restoran *new, *first = NULL, *firstnew, *tmp, *tmp2;

        /* Ucitavanje elemenata i poziv funkcije unesi. */
    for (i = 0; i < n; ++i) {
        new = (restoran*) malloc(sizeof(restoran));
        printf("Unesite naziv %d. restorana: ", i+1);
        scanf(" %[^\n]", new->naziv);
        printf("Unesite kucni broj (prirodan broj) i zaradu (realan broj) restorana %s: ", new->naziv);
        scanf("%d %lf", &new->kucni_broj, &new->zarada);
        first = unesi(first, new);
    }

        /* Kontrolni ispis polazne (ucitane) liste. */
    printf("\nPolazna lista:\n");
    ispis(first);

        /* Poziv funkcije prvih_k za zadani k > 0. */
    firstnew = prvih_k(&first, k);

        /* Kontrolni ispis izbacene podliste i ostatka ulazne liste. */
    printf("\nIzbacena podlista:\n");
    ispis(firstnew);
    printf("\nPreostala lista:\n");
    ispis(first);

        /* Dealokacija. */
    for (tmp = first; tmp != NULL; tmp = tmp2) {
        tmp2 = tmp->next;
        free(tmp);
    }
    for (tmp = firstnew; tmp != NULL; tmp = tmp2) {
        tmp2 = tmp->next;
        free(tmp);
    }

    return 0;
}
