/* Zadatak 3C:

Podatke o putovanjima u turisti\v{c}koj agenciji spremamo u vezanu listu.
Za svako putovanje pamtimo naziv (string od najvie 50 znakova),
broj prijavljenih (cijeli broj) i cijenu putovanja (realni broj).
Vezana lista putovanja je uvijek \textit{silazno\/} sortirana po broju prijavljenih.
Moete pretpostaviti da ne postoje dva putovanja s istim brojem prijavljenih.
\begin{itemize}
  \item[(a)]
	Napiite definiciju odgovarajuceg tipa podatka za putovanje, na nacin
	da bude moguce definirati varijable naredbom \texttt{put a;} i sl.
	Definirajte samo podatke koji su nuni za cuvanje takve liste u memoriji.
  \item[(b)]
	Napiite funkciju \texttt{put* unesi(put* first, put* new)} koja
	prima vezanu listu putovanja (\texttt{first} je pokazivac na pocetak
	ulazne vezane liste) te putovanje \texttt{new}. Funkcija ubacuje element
	\texttt{new} u vezanu listu tako da vezana lista i dalje bude
	\textit{silazno\/} sortirana po broju prijavljenih. Ulazna vezana lista moe
	biti prazna. Funkcija vraca pokazivac na prvi element dobivene vezane
	liste, nakon ubacivanja novog elementa.
  \item[(c)]
	Napiite funkciju \texttt{put* prvih\_k(put** first, int k)} koja
	trai prvih \texttt{k} susjednih putovanja s ukupnom najvecom cijenom
	i izbacuje ih iz vezane liste. Dva putovanja su susjedna ako u vezanoj
	listi ne postoji putovanje koje se nalazi izmedu njih.
	Funkcija vraca pokazivac na pocetak vezane liste izbacenih putovanja
	(koja, takoder, mora biti sortirana \textit{silazno\/} po broju
	prijavljenih). Ukoliko u ulaznoj vezanoj listi ima strogo manje od
	\texttt{k} elemenata, funkcija ne mijenja ulaznu vezanu listu i vraca
	\texttt{NULL}\@.
  \item[(d)]
	U glavnom programu formirajte vezanu listu od $5$ putovanja (elemenata),
	koristeci funkciju \texttt{unesi}. Podatke za ta putovanja mo\v{z}ete
	u\v{c}itati, ili zadati kao neke konstantne vrijednosti.
	Zatim, pozovite funkciju \texttt{prvih\_k} za $k = 4$. Na kraju,
	dealocirajte memoriju za sve elemente.
	Za rjeavanje ovog podzadatka nije nuno rijeiti podzadatke (b) i (c).
\end{itemize}
U (b) i (c) nije dozvoljeno alocirati memoriju za nove elemente, vec sve treba
raditi s postojecim elementima.

*/

#include <stdio.h>
#include <stdlib.h>

    /* Deklaracija tipa iz (a). */

typedef struct _put {
    char naziv[51];
    int br_prijava;
    double cijena;
    struct _put *next;
} put;

    /* Funkcija unesi iz (b). */

put* unesi(put* first, put* new)
{
    put *tmp;

	/* Slucaj kad je ulazna lista prazna. */
    if (first == NULL) {
        new->next = NULL;
        return new;
    }

        /* Slucaj kad novi element dodajemo na pocetak
           pa se mijenja prvi element liste. */
    if (new->br_prijava > first->br_prijava) {
        new->next = first;
        return new;
    }

        /* Ostali slucajevi = novi element dolazi iza
           nekog elementa u listi (first se NE mijenja). */
        /* Prazna for petlja koja staje u trenutku kad
           new treba doci izmedju tmp i tmp->next. */
    for (tmp = first; tmp->next != NULL && new->br_prijava < tmp->next->br_prijava; tmp = tmp->next);
    new->next = tmp->next;
    tmp->next = new;

    return first;
}

    /* Funkcija prvih_k iz (c). */
    /* Napomena: Pretpostavljamo da je k >= 1 (duljina trazenog bloka)! */

put* prvih_k(put** first, int k)
{
        /* Pokazivac na element ispred maksimalnih k, osim ako je pocetak
           tog bloka bas na pocetku liste, tj. nema onog ispred bloka. */
    put *maxfirst;
        /* Pokazivac na zadnji element u bloku od maksimalnih k. */
    put *maxlast;
        /* Analogni pokazivaci na blok od k elemenata kojeg provjeravamo,
           ako u listi ima vise takvih blokova (lista ima > k elemenata). */
    put *tmpfirst, *tmplast;

    int n = 0;
    int firstmax;  // Logicka varijabla = je li max. blok na pocetku liste.
    double maxcijena = 0.0, tmpcijena;

	/* Prvih k elemenata u listi uzimamo kao pocetni maksimum i
           pazimo da zaista imamo barem k elemenata u ulaznoj listi. */
    maxfirst = *first;
    for (maxlast = maxfirst; maxlast != NULL; maxlast = maxlast->next) {
        maxcijena += maxlast->cijena;
        ++n;
        if (n == k)
            break;    // Lista ima (barem) k elemenata.
    }

	/* Ako ulazna lista ima strogo manje od k elemenata,
           onda nista ne mijenjamo i vratimo NULL. */
    if (n < k)
        return NULL;

        /* Ulazna lista ima barem k elemenata, a maxlast pokazuje bas na
           k-ti element u listi (zbog break, maxlast se nije pomaknuo).
           Trenutni max. blok od k elemenata je na pocetku liste. */
    firstmax = 1;

        /* Provjeri ostale blokove od susjednih k elemenata (ako ih ima). */
        /* Pokazivac tmpfirst pokazuje na element ispred bloka od k susjednih
           elemenata, a tmplast pokazuje na zadnji element u tom bloku. */

    tmpfirst = maxfirst;        // Ispred sljedeceg bloka, ako ga ima.
    tmplast = maxlast->next;    // Zadnji u sljedecem bloku, ako ga ima.
    tmpcijena = maxcijena;

    while (tmplast != NULL) {
            /* Popravak cijene za sljedeci blok od k elemenata. */
        tmpcijena += tmplast->cijena - tmpfirst->cijena;
            /* Ako je to nova maksimalna, popravi sve podatke. */
        if (tmpcijena > maxcijena) {
            maxcijena = tmpcijena;
            maxfirst = tmpfirst;
            maxlast = tmplast;
            firstmax = 0;    // Max. blok vise nije na pocetku liste.
        }
        tmpfirst = tmpfirst->next;
        tmplast = tmplast->next;
    }

    if (firstmax) {
            /* Izbaci blok s pocetka liste. */
        *first = maxlast->next;
        maxlast->next = NULL;
        return maxfirst;
    }
    else {
            /* Izbaci blok iza maxfirst, do maxlast (ukljucivo). */
        tmpfirst = maxfirst->next;    // Zapamti pocetak izbacenog dijela.
        maxfirst->next = maxlast->next;
        maxlast->next = NULL;         // Kraj izbacenog dijela.
        return tmpfirst;
    }
}

    /* Pomocna funkcija za ispis elemenata liste. */

void ispis(put *first)
{
    put *tmp;

    for (tmp = first; tmp != NULL; tmp = tmp->next) {
        printf("%s\n%d, %f\n\n", tmp->naziv, tmp->br_prijava, tmp->cijena);
    }
    printf("\n");

    return;
}


    /* Glavni program iz (d). */

int main(void)
{
    int n = 5, k = 4, i;
    put *new, *first = NULL, *firstnew, *tmp, *tmp2;

        /* Ucitavanje elemenata i poziv funkcije unesi. */
    for (i = 0; i < n; ++i) {
        new = (put*) malloc(sizeof(put));
        printf("Unesite naziv %d. puta: ", i+1);
        scanf(" %[^\n]", new->naziv);
        printf("Unesite broj prijavljenih (prirodan broj) i cijenu (realan broj) puta %s: ", new->naziv);
        scanf("%d %lf", &new->br_prijava, &new->cijena);
        first = unesi(first, new);
    }

        /* Kontrolni ispis polazne (ucitane) liste. */
    printf("\nPolazna lista:\n");
    ispis(first);

        /* Poziv funkcije prvih_k za zadani k > 0. */
    firstnew = prvih_k(&first, k);

        /* Kontrolni ispis izbacene podliste i ostatka ulazne liste. */
    printf("\nIzbacena podlista:\n");
    ispis(firstnew);
    printf("\nPreostala lista:\n");
    ispis(first);

        /* Dealokacija. */
    for (tmp = first; tmp != NULL; tmp = tmp2) {
        tmp2 = tmp->next;
        free(tmp);
    }
    for (tmp = firstnew; tmp != NULL; tmp = tmp2) {
        tmp2 = tmp->next;
        free(tmp);
    }

    return 0;
}
