/* Zadatak 3D:

Podatke o lijekovima u jednoj apoteci spremamo u vezanu listu. Za svaki lijek
pamtimo naziv (string od najvie 50 znakova), broj kupaca (prirodni broj) i
koli\v{c}inu (realni broj).
Vezana lista lijekova je uvijek \textit{uzlazno\/} sortirana po broju kupaca.
Moete pretpostaviti da ne postoje dva lijeka s istim brojem kupaca.
\begin{itemize}
  \item[(a)]
	Napiite definiciju odgovarajuceg tipa podatka za lijek, na nacin
	da bude moguce definirati varijable naredbom \texttt{lijek~a;} i sl.
	Definirajte samo podatke koji su nuni za cuvanje takve liste u memoriji.
  \item[(b)]
	Napiite funkciju \texttt{lijek* unesi(lijek* first, lijek* new)} koja
	prima vezanu listu lijekova (\texttt{first} je pokazivac na pocetak
	ulazne vezane liste) te lijek \texttt{new}. Funkcija ubacuje element
	\texttt{new} u vezanu listu tako da vezana lista i dalje bude
	\textit{uzlazno\/} sortirana po broju kupaca. Ulazna vezana lista moe
	biti prazna. Funkcija vraca pokazivac na prvi element dobivene vezane
	liste, nakon ubacivanja novog elementa.
  \item[(c)]
	Napiite funkciju \texttt{lijek* prvih\_k(lijek** first, int k)} koja
	trai prvih \texttt{k} susjednih lijekova s ukupnom najmanjom koli\v{c}inom
	i izbacuje ih iz vezane liste. Dva lijeka su susjedna ako u vezanoj
	listi ne postoji lijek koji se nalazi izmedu njih.
	Funkcija vraca pokazivac na pocetak vezane liste izbacenih lijekova
	(koja, takoder, mora biti sortirana \textit{uzlazno\/} po broju
	kupaca). Ukoliko u ulaznoj vezanoj listi ima strogo manje od
	\texttt{k} elemenata, funkcija ne mijenja ulaznu vezanu listu i vraca
	\texttt{NULL}\@.
  \item[(d)]
	U glavnom programu formirajte vezanu listu od $5$ lijekova (elemenata),
	koristeci funkciju \texttt{unesi}. Podatke za te lijekove mo\v{z}ete
	u\v{c}itati, ili zadati kao neke konstantne vrijednosti.
	Zatim, pozovite funkciju \texttt{prvih\_k} za $k = 1$. Na kraju,
	dealocirajte memoriju za sve elemente.
	Za rjeavanje ovog podzadatka nije nuno rijeiti podzadatke (b) i (c).
\end{itemize}
U (b) i (c) nije dozvoljeno alocirati memoriju za nove elemente, vec sve treba
raditi s postojecim elementima.

*/

#include <stdio.h>
#include <stdlib.h>

    /* Deklaracija tipa iz (a). */

typedef struct _lijek {
    char naziv[51];
    int br_kupaca;
    double kolicina;
    struct _lijek *next;
} lijek;

    /* Funkcija unesi iz (b). */

lijek* unesi(lijek* first, lijek* new)
{
    lijek *tmp;

	/* Slucaj kad je ulazna lista prazna. */
    if (first == NULL) {
        new->next = NULL;
        return new;
    }

        /* Slucaj kad novi element dodajemo na pocetak
           pa se mijenja prvi element liste. */
    if (new->br_kupaca < first->br_kupaca) {
        new->next = first;
        return new;
    }

        /* Ostali slucajevi = novi element dolazi iza
           nekog elementa u listi (first se NE mijenja). */
        /* Prazna for petlja koja staje u trenutku kad
           new treba doci izmedju tmp i tmp->next. */
    for (tmp = first; tmp->next != NULL && new->br_kupaca > tmp->next->br_kupaca; tmp = tmp->next);
    new->next = tmp->next;
    tmp->next = new;

    return first;
}

    /* Funkcija prvih_k iz (c). */
    /* Napomena: Pretpostavljamo da je k >= 1 (duljina trazenog bloka)! */

lijek* prvih_k(lijek** first, int k)
{
        /* Pokazivac na element ispred minimalnih k, osim ako je pocetak
           tog bloka bas na pocetku liste, tj. nema onog ispred bloka. */
    lijek *minfirst;
        /* Pokazivac na zadnji element u bloku od minimalnih k. */
    lijek *minlast;
        /* Analogni pokazivaci na blok od k elemenata kojeg provjeravamo,
           ako u listi ima vise takvih blokova (lista ima > k elemenata). */
    lijek *tmpfirst, *tmplast;

    int n = 0;
    int firstmin;  // Logicka varijabla = je li min. blok na pocetku liste.
    double minkolicina = 0.0, tmpkolicina;

	/* Prvih k elemenata u listi uzimamo kao pocetni minimum i
           pazimo da zaista imamo barem k elemenata u ulaznoj listi. */
    minfirst = *first;
    for (minlast = minfirst; minlast != NULL; minlast = minlast->next) {
        minkolicina += minlast->kolicina;
        ++n;
        if (n == k)
            break;    // Lista ima (barem) k elemenata.
    }

	/* Ako ulazna lista ima strogo manje od k elemenata,
           onda nista ne mijenjamo i vratimo NULL. */
    if (n < k)
        return NULL;

        /* Ulazna lista ima barem k elemenata, a minlast pokazuje bas na
           k-ti element u listi (zbog break, minlast se nije pomaknuo).
           Trenutni min. blok od k elemenata je na pocetku liste. */
    firstmin = 1;

        /* Provjeri ostale blokove od susjednih k elemenata (ako ih ima). */
        /* Pokazivac tmpfirst pokazuje na element ispred bloka od k susjednih
           elemenata, a tmplast pokazuje na zadnji element u tom bloku. */

    tmpfirst = minfirst;        // Ispred sljedeceg bloka, ako ga ima.
    tmplast = minlast->next;    // Zadnji u sljedecem bloku, ako ga ima.
    tmpkolicina = minkolicina;

    while (tmplast != NULL) {
            /* Popravak kolicine za sljedeci blok od k elemenata. */
        tmpkolicina += tmplast->kolicina - tmpfirst->kolicina;
            /* Ako je to nova minimalna, popravi sve podatke. */
        if (tmpkolicina < minkolicina) {
            minkolicina = tmpkolicina;
            minfirst = tmpfirst;
            minlast = tmplast;
            firstmin = 0;    // Min. blok vise nije na pocetku liste.
        }
        tmpfirst = tmpfirst->next;
        tmplast = tmplast->next;
    }

    if (firstmin) {
            /* Izbaci blok s pocetka liste. */
        *first = minlast->next;
        minlast->next = NULL;
        return minfirst;
    }
    else {
            /* Izbaci blok iza minfirst, do minlast (ukljucivo). */
        tmpfirst = minfirst->next;    // Zapamti pocetak izbacenog dijela.
        minfirst->next = minlast->next;
        minlast->next = NULL;         // Kraj izbacenog dijela.
        return tmpfirst;
    }
}

    /* Pomocna funkcija za ispis elemenata liste. */

void ispis(lijek *first)
{
    lijek *tmp;

    for (tmp = first; tmp != NULL; tmp = tmp->next) {
        printf("%s\n%d, %f\n\n", tmp->naziv, tmp->br_kupaca, tmp->kolicina);
    }
    printf("\n");

    return;
}


    /* Glavni program iz (d). */

int main(void)
{
    int n = 5, k = 1, i;
    lijek *new, *first = NULL, *firstnew, *tmp, *tmp2;

        /* Ucitavanje elemenata i poziv funkcije unesi. */
    for (i = 0; i < n; ++i) {
        new = (lijek*) malloc(sizeof(lijek));
        printf("Unesite naziv %d. lijeka: ", i+1);
        scanf(" %[^\n]", new->naziv);
        printf("Unesite broj kupaca (prirodan broj) i kolicinu (realan broj) lijeka %s: ", new->naziv);
        scanf("%d %lf", &new->br_kupaca, &new->kolicina);
        first = unesi(first, new);
    }

        /* Kontrolni ispis polazne (ucitane) liste. */
    printf("\nPolazna lista:\n");
    ispis(first);

        /* Poziv funkcije prvih_k za zadani k > 0. */
    firstnew = prvih_k(&first, k);

        /* Kontrolni ispis izbacene podliste i ostatka ulazne liste. */
    printf("\nIzbacena podlista:\n");
    ispis(firstnew);
    printf("\nPreostala lista:\n");
    ispis(first);

        /* Dealokacija. */
    for (tmp = first; tmp != NULL; tmp = tmp2) {
        tmp2 = tmp->next;
        free(tmp);
    }
    for (tmp = firstnew; tmp != NULL; tmp = tmp2) {
        tmp2 = tmp->next;
        free(tmp);
    }

    return 0;
}
